# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
STL element
https://en.wikipedia.org/wiki/STL_(file_format)
"""

# pylint: disable=invalid-name

from os.path import isfile
import logging
# from textwrap import dedent
import numpy as np
from stl import mesh

from .base import Element
# from .utils import distance_lines

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def intersection(grid, v, tri):
    """
    We look for the intersection between
    - the lines with the direction v that passes by x, y, z
    - the triangle tri

    t>0, a>=0, b>=0, a+b<=1
    x + t vx = a tri[6] + b tri[3] + (1-a-b) tri[0]
    y + t vy = a tri[7] + b tri[4] + (1-a-b) tri[1]
    z + t vz = a tri[8] + b tri[5] + (1-a-b) tri[2]
    <=>
    a (tri[6]-tri[0]) + b (tri[3]-tri[0]) - t vx = x-tri[0]
    a (tri[7]-tri[1]) + b (tri[4]-tri[1]) - t vy = y-tri[1]
    a (tri[8]-tri[2]) + b (tri[5]-tri[2]) - t vz = z-tri[2]
    """
    # the matrix of the intersection problem
    # A is 3x3 and does not depend on the grid
    A = np.array(
        [
            [tri[6]-tri[0], tri[3]-tri[0], -v[0]],
            [tri[7]-tri[1], tri[4]-tri[1], -v[1]],
            [tri[8]-tri[2], tri[5]-tri[2], -v[2]]
        ]
    )
    # if the direction v is in the planed of the triangle tri
    # then the matrix A is not invertible
    # the intersection is not computed with that v but with another one
    try:
        invA = np.linalg.inv(A)
    except np.linalg.LinAlgError as e:
        if 'Singular matrix' not in str(e):
            raise
        return None

    # right member of the intersection problem
    # it depends on the grid and is a (xsize, ysize, zsize) vector
    x, y, z = grid
    P = np.asarray(
        [x - tri[0], y - tri[1], z - tri[2]],
        dtype=object
    )
    # solve the problem
    sol = invA.dot(P)
    # test if the intersection point is
    # - inside the triangle: s is computed
    # - outside the triangle: s = np.inf
    indin = np.logical_and(
        np.logical_and(
            sol[0] >= 0, sol[1] >= 0
        ), np.logical_and(
            sol[0] + sol[1] <= 1, sol[2] >= 0
        )
    )
    s = np.full((x.size, y.size, z.size), np.inf)
    s[indin] = sol[2][indin]
    # # compute the coordinates of the intersection point
    # p = np.full((3, x.size, y.size, z.size), np.inf)
    # for k in range(3):
    #     p[k][indin] = tri[3*k] * sol[0][indin] + \
    #         tri[3*k+1] * sol[1][indin] + \
    #         tri[3*k+2] * (1-sol[0][indin]-sol[1][indin])

    return s


class STLElement(Element):
    """
    Class STLElement

    Notes
    --------
    Add a STLElement requires the module numpy-stl!

    Parameters
    ----------

    filename: string
        the file containing the STL element
    label: int
        the label on the element
    isfluid: boolean
        - True if the element is added
        - False if the element is deleted

    Examples
    --------

    Attributes
    ----------

    mesh: stl.mesh.Mesh
        the surface mesh of the element
    dim: int
        3
    """
    def __init__(self, filename, label=0, isfluid=False):
        self.filename = filename
        self.mesh = mesh.Mesh.from_file(filename)
        self.number_of_bounds = 1  # just one bound for the labels
        self.nb_tri = self.mesh.points.shape[0]  # nb of triangles
        self.dim = 3
        super(STLElement, self).__init__(label, isfluid)
        log.info(self.__str__())

    def get_bounds(self):
        return self.mesh.min_, self.mesh.max_

    def _center(self):
        return self.mesh.get_mass_properties()[1]

    def point_inside(self, grid):
        shape = tuple([gk.size for gk in grid])
        in_or_out = np.full(shape, False, dtype=bool)
        smin = np.full(shape, np.inf)

        # for each spatial direction, we plot a line
        # and we compute the distance of the first intersection
        # with the element in that direction.
        # if the normal is in the opposite side, the point is outside
        for direction in range(self.dim):
            smin[:] = np.inf
            v = np.zeros((self.dim,))
            v[direction] = 1
            for ind_tri in range(self.nb_tri):
                tri = self.mesh.points[ind_tri, :]
                s = intersection(grid, v, tri)
                if s is not None:
                    indices = np.asarray(s < smin).nonzero()
                    smin[indices] = s[indices]
                    if np.inner(v, self.mesh.normals[ind_tri]) >= 0:
                        in_or_out[indices] = True
                    else:
                        in_or_out[indices] = False
        return in_or_out

    def distance(self, grid, v, dmax=None, normal=False):
        shape = tuple([gk.size for gk in grid])
        alpha = np.full(shape, np.inf)
        border = np.full(shape, -1)
        if normal:
            normal_x = np.zeros(shape)
            normal_y = np.zeros(shape)
            normal_z = np.zeros(shape)
        # loop over the triangles
        for ind_tri in range(self.nb_tri):
            tri = self.mesh.points[ind_tri, :]
            s = intersection(grid, v, tri)
            if s is not None:
                if dmax is None:
                    indices = np.asarray(s < alpha).nonzero()
                else:
                    indices = np.logical_and(
                        s < alpha, s <= dmax
                    )
                alpha[indices] = s[indices]
                border[indices] = self.label[0]
                if normal:
                    normal_x[indices] = self.mesh.normals[ind_tri][0]
                    normal_y[indices] = self.mesh.normals[ind_tri][1]
                    normal_z[indices] = self.mesh.normals[ind_tri][2]
        if normal:
            normal = np.zeros(tuple(list(shape) + [3]))
            normal[..., 0] = normal_x
            normal[..., 1] = normal_y
            normal[..., 2] = normal_z
        return alpha, border, normal

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('stl.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        elem_dim = np.asarray(self.get_bounds()).reshape((6,))
        return template.render(
            header=header_string(self.__class__.__name__),
            elem=self, type=elem_type, dim=elem_dim
        )

    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(3), alpha=0.25,
                  ):
        # coordinates of the basis triangle
        p = np.asarray(
            [
                [0, 0, 0],
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 0],
            ]
        ).T
        vv = np.array([0, 1])
        # meshgrid used for plot_surface
        Xb = np.outer(p[0], vv)
        Yb = np.outer(p[1], vv)
        Zb = np.outer(p[2], vv)

        points = self.mesh.points
        normals = self.mesh.normals
        for k in range(points.shape[0]):
            # plot the kth triangle
            tri = points[k]
            n = normals[k]
            # the coordinates of the three points
            x, y, z = tri[::3], tri[1::3], tri[2::3]
            # matrix of the change of basis
            A = np.array(
                [
                    [x[1]-x[0], x[2]-x[0], n[0]],
                    [y[1]-y[0], y[2]-y[0], n[1]],
                    [z[1]-z[0], z[2]-z[0], n[2]]
                ]
            )
            # write the coordinates of the triangle
            X = x[0] + A[0, 0]*Xb + A[0, 1]*Yb + A[0, 2]*Zb
            Y = y[0] + A[1, 0]*Xb + A[1, 1]*Yb + A[1, 2]*Zb
            Z = z[0] + A[2, 0]*Xb + A[2, 1]*Yb + A[2, 2]*Zb
            # plot
            viewer.surface(X, Y, Z, 'black', alpha=alpha)

        if viewlabel:
            x, y, z = self._center()
            viewer.text(str(self.label[0]), [x, y, z])



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Ellipse element
"""
# pylint: disable=invalid-name

import logging
# from textwrap import dedent
import numpy as np

from .base import Element
from .utils import distance_ellipse

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Ellipse(Element):
    """
    Class Ellipse

    Parameters
    ----------
    center : list
        the two coordinates of the center
    v1 : list
        a vector
    v2 : list
        a second vector (v1 and v2 have to be othogonal)
    label : list
        one integer (default [0])
    isfluid : boolean
        - True if the ellipse is added
        - False if the ellipse is deleted

    Attributes
    ----------
    number_of_bounds : int
        1
    dim: int
        2
    center : ndarray
        the coordinates of the center of the ellipse
    v1 : ndarray
        the coordinates of the first vector
    v2 : ndarray
        the coordinates of the second vector
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the ellipse is added
        and False if the ellipse is deleted

    Examples
    --------

    the ellipse centered in (0, 0) with v1=[2,0], v2=[0,1]

    >>> center = [0., 0.]
    >>> v1 = [2., 0.]
    >>> v2 = [0., 1.]
    >>> Ellipse(center, v1, v2)
    +---------+
    | Ellipse |
    +---------+
        - dimension: 2
        - center: [0. 0.]
        - v1: [2. 0.]
        - v2: [0. 1.]
        - label: [0]
        - type: solid

    """
    def __init__(self, center, v1, v2, label=0, isfluid=False):
        self.number_of_bounds = 1  # number of edges
        self.dim = 2
        self.center = np.asarray(center)
        if abs(v1[0]*v2[0] + v1[1]*v2[1]) > 1.e-14:
            log.error('The vectors of the ellipse are not orthogonal')
        else:
            self.v1 = np.asarray(v1)
            self.v2 = np.asarray(v2)
        super(Ellipse, self).__init__(label, isfluid)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the ellipse.
        """
        r = max(np.linalg.norm(self.v1), np.linalg.norm(self.v2))
        return self.center - r, self.center + r

    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the ellipse.

        Notes
        -----

        the edge of the ellipse is considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the ellipse, False otherwise)

        """
        x, y = grid

        X = x - self.center[0]
        Y = y - self.center[1]
        vx2 = self.v1[0]**2 + self.v2[0]**2
        vy2 = self.v1[1]**2 + self.v2[1]**2
        vxy = 2 * (self.v1[0]*self.v1[1] + self.v2[0]*self.v2[1])
        tv = self.v1[0]*self.v2[1]-self.v1[1]*self.v2[0]
        return X**2*vy2 + Y**2*vx2 - X*Y*vxy <= tv**2

    def distance(self, grid, v, dmax=None, normal=False):
        """
        Compute the distance in the v direction between
        the ellipse and the points defined by (x, y).

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max
        normal : bool
            return the normal vector if True (default False)

        Returns
        -------

        ndarray
            array of distances if normal is False and
            the coordinates of the normal vectors
            if normal is True
        """
        x, y = grid
        return distance_ellipse(
            x, y, v,
            self.center, self.v1, self.v2,
            dmax, self.label, normal
        )

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('ellipse.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        return template.render(
            header=header_string(self.__class__.__name__),
            elem=self,
            type=elem_type
        )

    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(2), alpha=1.
                  ):
        nv1 = np.linalg.norm(self.v1)
        nv2 = np.linalg.norm(self.v2)
        if nv1 > nv2:
            r1, r2 = nv1, nv2
            v = self.v1
        else:
            r1, r2 = nv2, nv2
            v = self.v2
        if v[0] == 0:
            theta = .5*np.pi
        else:
            theta = np.arctan(v[1]/v[0])

        viewer.ellipse(
            self.center*scale, (r1*scale[0], r2*scale[1]),
            color, angle=theta, alpha=alpha
        )
        if viewlabel:
            x = self.center[0] + r1*np.cos(theta)
            y = self.center[1] + r1*np.sin(theta)
            viewer.text(str(self.label[0]), [x*scale[0], y*scale[1]])



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Circle element
"""

# pylint: disable=invalid-name

import logging
# from textwrap import dedent
import numpy as np

from .base import Element
from .utils import distance_ellipse

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Circle(Element):
    """
    Class Circle

    Parameters
    ----------
    center : list
        the two coordinates of the center
    radius : float
        the radius
    label : list
        default [0]
    isfluid : boolean
        - True if the circle is added
        - False if the circle is deleted

    Attributes
    ----------
    number_of_bounds : int
        1
    dimension: int
        2
    center : ndarray
        the coordinates of the center of the circle
    radius : double
        positive float for the radius of the circle
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the circle is added
        and False if the circle is deleted

    Examples
    --------

    the circle centered in (0, 0) with radius 1

    >>> center = [0., 0.]
    >>> radius = 1.
    >>> Circle(center, radius)
    +--------+
    | Circle |
    +--------+
        - dimension: 2
        - center: [0. 0.]
        - radius: 1.0
        - label: [0]
        - type: solid

    """
    def __init__(self, center, radius, label=0, isfluid=False):
        self.number_of_bounds = 1  # number of edges
        self.dim = 2
        self.center = np.asarray(center)
        if radius >= 0:
            self.radius = radius
        else:
            log.error('The radius of the circle should be positive')
        super(Circle, self).__init__(label, isfluid)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the circle.
        """
        return self.center - self.radius, self.center + self.radius

    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the circle.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the circle, False otherwise)

        Notes
        -----

        the edge of the circle is considered as inside.

        """
        x, y = grid
        v2 = np.asarray([x - self.center[0], y - self.center[1]], dtype=object)
        return (v2[0]**2 + v2[1]**2) <= self.radius**2

    def distance(self, grid, v, dmax=None, normal=False):
        """
        Compute the distance in the v direction between
        the circle and the points defined by (x, y).
        if normal==True, compute also the normal vector

        .. image:: ../figures/Circle.png
            :width: 100%

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max (default None)
        normal : bool
            return the normal vector if True (default False)

        Returns
        -------

        ndarray
            array of distances if normal is False and
            the coordinates of the normal vectors
            if normal is True
        """
        x, y = grid
        v1 = self.radius*np.array([1, 0])
        v2 = self.radius*np.array([0, 1])
        return distance_ellipse(
            x, y, v, self.center, v1, v2,
            dmax, self.label, normal
        )

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('circle.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        return template.render(
            header=header_string('Circle'),
            elem=self, type=elem_type
        )

    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(2), alpha=1.
                  ):
        viewer.ellipse(self.center*scale,
                       tuple(self.radius*scale),
                       color,
                       alpha=alpha
                       )
        if viewlabel:
            theta = self.center[0] + 2*self.center[1]+10*self.radius
            x = self.center[0] + self.radius*np.cos(theta)
            y = self.center[1] + self.radius*np.sin(theta)
            viewer.text(str(self.label[0]), [x*scale[0], y*scale[1]])



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

from .base import BaseAlgorithm

class PullAlgorithm(BaseAlgorithm):
    def one_time_step_local(self, f, fnew, m):
        """
        Return symbolic expression which makes one time step of
        LBM algorithm using the pull algorithm. The difference with
        the basic algorithm is the transport and the computation
        of the moments from the distributed functions is made in
        one step.

            - compute the moments from the distributed functions + transport
            - source terms with dt/2 (with the moments)
            - relaxation (with the moments)
            - source terms with dt/2 (with the moments)
            - compute the new distributed functions from the moments

        Parameters
        ----------

        f : SymPy Matrix
            indexed objects for the old distributed functions

        fnew : SymPy Matrix
            indexed objects for the new distributed functions

        m : SymPy Matrix
            indexed objects for the moments

        """
        with_rel_velocity = True if self.rel_vel_symb else False

        f2m = self.f2m_local(f, m, with_rel_velocity)
        if isinstance(f2m, list):
            code = f2m
        else:
            code = [f2m]

        if self.source_eq:
            code.extend(self.source_term_local(m))

        code.append(self.relaxation_local(m, with_rel_velocity))

        if with_rel_velocity:
            code.append(self.restore_conserved_moments(m, f))

        if self.source_eq:
            code.extend(self.source_term_local(m))

        code.append(self.m2f_local(m, fnew, with_rel_velocity))

        return code



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Cylinder element with
    - circle base
    - ellipse base
    - triangle base
    - parallelogram base
"""

# pylint: disable=invalid-name, no-member, attribute-defined-outside-init
# pylint: disable=wildcard-import, unused-wildcard-import

import logging
# from textwrap import dedent
import numpy as np

from .base import *  # pylint: disable=redefined-builtin

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Cylinder(Element):
    """
    Class Cylinder

    generic class for the cylinders
    """
    dim = 3

    def change_of_variables(self):
        """
        matrix for the change of variables
        used to write the coordinates in the basis of the cylinder

        After the change of variables, 
        the base of the cylinder is in the plane x, y
        and the other direction is limited by -1 < z < 1
        """
        self.A = np.empty((3, 3))
        self.A[:, 0] = self.v1
        self.A[:, 1] = self.v2
        self.A[:, 2] = self.w
        self.iA = np.linalg.inv(self.A)

    def get_bounds(self):
        """
        Get the bounds of the cylinder.

        Returns
        -------

        ndarray
            minimal box where the cylinder is included

        """
        lw = [0.5*abs(self.w[k]) for k in range(len(self.w))]
        bounds_base = self.base.get_bounds()
        return bounds_base[0] - lw, bounds_base[1] + lw

    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the cylinder.

        Notes
        -----

        the edge of the cylinder is considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the cylinder, False otherwise)

        """
        x, y, z = grid
        xx = x - self.center[0]
        yy = y - self.center[1]
        zz = z - self.center[2]
        # the new x coordinates
        x_cyl = self.iA[0, 0]*xx + self.iA[0, 1]*yy + self.iA[0, 2]*zz
        # the new y coordinates
        y_cyl = self.iA[1, 0]*xx + self.iA[1, 1]*yy + self.iA[1, 2]*zz
        # the new z coordinates
        z_cyl = self.iA[2, 0]*xx + self.iA[2, 1]*yy + self.iA[2, 2]*zz
        return np.logical_and(
            self.base.point_inside((x_cyl, y_cyl)),
            np.abs(z_cyl) <= 1.
        )

    # pylint: disable=too-many-locals
    def distance(self, grid, v, dmax=None, normal=False):
        """
        Compute the distance in the v direction between
        the cylinder and the points defined by (x, y, z).

        .. image:: ../figures/Cylinder.png
            :width: 100%

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max
        normal : bool
            return the normal vector if True (default False)

        Returns
        -------

        ndarray
            array of distances if normal is False and
            the coordinates of the normal vectors
            if normal is True
        """
        x, y, z = grid

        # rewritte the coordinates in the frame of the cylinder
        v_cyl = self.iA.dot(np.asarray(v))  # the velocity
        xx = x - self.center[0]
        yy = y - self.center[1]
        zz = z - self.center[2]
        # the new x coordinates
        x_cyl = self.iA[0, 0]*xx + self.iA[0, 1]*yy + self.iA[0, 2]*zz
        # the new y coordinates
        y_cyl = self.iA[1, 0]*xx + self.iA[1, 1]*yy + self.iA[1, 2]*zz
        # the new z coordinates
        z_cyl = self.iA[2, 0]*xx + self.iA[2, 1]*yy + self.iA[2, 2]*zz

        # normal vector in the frame of the cylinder
        normal_cyl = np.zeros(tuple(list(x_cyl.shape) + [3]))
        normal_cyl_x = normal_cyl[..., 0]
        normal_cyl_y = normal_cyl[..., 1]
        normal_cyl_z = normal_cyl[..., 2]
        # considering the infinite cylinder
        alpha, border, normal_cyl[..., :2] = self.base.distance(
            (x_cyl, y_cyl),
            v_cyl[:-1],
            dmax, self.label[:-2], normal
        )
        # indices where the intersection is too high or to low
        alpha[alpha < 0] = 1.e16
        border[alpha < 0] = -1
        normal_cyl_x[alpha < 0] = 0
        normal_cyl_y[alpha < 0] = 0
        normal_cyl_z[alpha < 0] = 0
        ind = np.logical_and(alpha > 0, np.abs(z_cyl + alpha*v_cyl[2]) > 1.)
        alpha[ind] = 1.e16
        border[ind] = -1.
        normal_cyl_x[ind] = 0
        normal_cyl_y[ind] = 0
        normal_cyl_z[ind] = 0

        # considering the two planes
        dummyf = self.base.point_inside
        if v_cyl[2] == 0:  # to avoid vertical velocities
            decal = 1.e-16
        else:
            decal = 0.
        alpha_top = (1.-z_cyl)/(v_cyl[2] + decal)
        ind = np.logical_or(
            np.logical_or(alpha_top < 0, alpha_top > dmax),
            np.logical_not(dummyf(
                (x_cyl + alpha_top*v_cyl[0], y_cyl + alpha_top*v_cyl[1])
            ))
        )
        alpha_top[ind] = 1.e16
        alpha_bot = -(1.+z_cyl)/(v_cyl[2] + decal)
        ind = np.logical_or(
            np.logical_or(alpha_bot < 0, alpha_bot > dmax),
            np.logical_not(dummyf(
                (x_cyl + alpha_bot*v_cyl[0], y_cyl + alpha_bot*v_cyl[1])
            ))
        )
        alpha_bot[ind] = 1.e16

        # considering the first intersection point
        alpha = np.amin([alpha, alpha_top, alpha_bot], axis=0)
        # fix the top
        ind = alpha == alpha_top
        border[ind] = self.label[-1]
        normal_cyl_x[ind] = 0.
        normal_cyl_y[ind] = 0.
        normal_cyl_z[ind] = 1.
        # fix the bottom
        ind = alpha == alpha_bot
        border[ind] = self.label[-2]
        normal_cyl_x[ind] = 0.
        normal_cyl_y[ind] = 0.
        normal_cyl_z[ind] = -1.
        # fix the distance
        alpha[alpha == 1.e16] = -1.
        # rewritte the normal vector in the initial frame
        normal_vect = np.zeros(normal_cyl.shape)
        for k in range(3):
            normal_vect[..., k] = self.A[k, 0] * normal_cyl_x + \
                self.A[k, 1] * normal_cyl_y + self.A[k, 2] * normal_cyl_z

        return alpha, border, normal_vect

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('cylinder.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        return template.render(header=header_string(self.__class__.__name__),
                               elem=self, type=elem_type)

    # pylint: disable=too-many-locals
    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(3), alpha=1.
                  ):
        if not isinstance(color, list):
            color = [color]*self.number_of_bounds
        lx_b, ly_b = self.base._visualize()  # pylint: disable=protected-access
        c = self.center
        for k in range(len(lx_b)-2):  # loop over the faces of the side
            x_b, y_b = lx_b[k], ly_b[k]
            z_b = [-1., 1.]
            X_cyl, Z_cyl = np.meshgrid(x_b, z_b)
            Y_cyl, Z_cyl = np.meshgrid(y_b, z_b)
            X = c[0] + self.A[0, 0]*X_cyl \
                + self.A[0, 1]*Y_cyl \
                + self.A[0, 2]*Z_cyl
            Y = c[1] + self.A[1, 0]*X_cyl \
                + self.A[1, 1]*Y_cyl \
                + self.A[1, 2]*Z_cyl
            Z = c[2] + self.A[2, 0]*X_cyl \
                + self.A[2, 1]*Y_cyl \
                + self.A[2, 2]*Z_cyl
            viewer.surface(X, Y, Z, color[k], alpha=alpha)
        vv = np.sin(np.linspace(0, np.pi, 10))
        Xbase = np.outer(lx_b[-2], vv)
        Ybase = np.outer(ly_b[-2], vv)
        Zbase = np.ones(Xbase.shape)
        X = c[0] + self.A[0, 0]*Xbase + self.A[0, 1]*Ybase + self.A[0, 2]*Zbase
        Y = c[1] + self.A[1, 0]*Xbase + self.A[1, 1]*Ybase + self.A[1, 2]*Zbase
        Z = c[2] + self.A[2, 0]*Xbase + self.A[2, 1]*Ybase + self.A[2, 2]*Zbase
        viewer.surface(X, Y, Z, color=color[-2], alpha=alpha)
        X = c[0] + self.A[0, 0]*Xbase + self.A[0, 1]*Ybase - self.A[0, 2]*Zbase
        Y = c[1] + self.A[1, 0]*Xbase + self.A[1, 1]*Ybase - self.A[1, 2]*Zbase
        Z = c[2] + self.A[2, 0]*Xbase + self.A[2, 1]*Ybase - self.A[2, 2]*Zbase
        viewer.surface(X, Y, Z, color=color[-2], alpha=alpha)


class CylinderCircle(Cylinder):
    """
    Class CylinderCircle

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the first vector that defines the circular section
    v2 : list
        the second vector that defines the circular section
    w : list
        the vector that defines the direction of the side
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Attributes
    ----------
    number_of_bounds : int
        3
    dim: int
        3
    center : ndarray
        the coordinates of the center of the cylinder
    v1 : list
        the three coordinates of the first vector defining the base section
    v2 : list
        the three coordinates of the second vector defining the base section
    w : list
        the three coordinates of the vector defining the direction of the side
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the cylinder is added
        and False if the cylinder is deleted

    Examples
    --------

    the vertical canonical cylinder centered in (0, 0, 1/2) with radius 1

    >>> center = [0., 0., 0.5]
    >>> v1, v2 = [1., 0., 0.], [0., 1., 0.]
    >>> w = [0., 0., 1.]
    >>> CylinderCircle(center, v1, v2, w)
    +----------------+
    | CylinderCircle |
    +----------------+
        - dimension: 3
        - center: [0.  0.  0.5]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0. 0. 1.]
        - label: [0, 0, 0]
        - type: solid

    """
    def __init__(self, center, v1, v2, w, label=0, isfluid=False):
        self.number_of_bounds = 3  # number of edges
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        self.w = np.asarray(w)
        self.change_of_variables()
        self.base = BaseCircle(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)


class CylinderEllipse(Cylinder):
    """
    Class CylinderEllipse

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the first vector that defines the circular section
    v2 : list
        the second vector that defines the circular section
    w : list
        the vector that defines the direction of the side
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Warnings
    --------

    The vectors v1 and v2 have to be orthogonal.

    Attributes
    ----------
    number_of_bounds : int
        3
    dim: int
        3
    center : ndarray
        the coordinates of the center of the cylinder
    v1 : list
        the three coordinates of the first vector defining the base section
    v2 : list
        the three coordinates of the second vector defining the base section
    w : list
        the three coordinates of the vector defining the direction of the side
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the cylinder is added
        and False if the cylinder is deleted

    Examples
    --------

    the vertical canonical cylinder centered in (0, 0, 1/2) with radius 1

    >>> center = [0., 0., 0.5]
    >>> v1, v2 = [1., 0., 0.], [0., 1., 0.]
    >>> w = [0., 0., 1.]
    >>> CylinderEllipse(center, v1, v2, w)
    +-----------------+
    | CylinderEllipse |
    +-----------------+
        - dimension: 3
        - center: [0.  0.  0.5]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0. 0. 1.]
        - label: [0, 0, 0]
        - type: solid

    """
    def __init__(self, center, v1, v2, w, label=0, isfluid=False):
        self.number_of_bounds = 3  # number of edges
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        self.w = np.asarray(w)
        self.change_of_variables()
        self.base = BaseEllipse(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)


class CylinderTriangle(Cylinder):
    """
    Class CylinderTriangle

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the first vector that defines the triangular section
    v2 : list
        the second vector that defines the triangular section
    w : list
        the vector that defines the direction of the side
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Attributes
    ----------
    number_of_bounds : int
        5
    dim: int
        3
    center : numpy array
        the coordinates of the center of the cylinder
    v1 : list of doubles
        the three coordinates of the first vector defining the base section
    v2 : list of doubles
        the three coordinates of the second vector defining the base section
    w : list of doubles
        the three coordinates of the vector defining the direction of the side
    label : list of integers
        the list of the label of the edge
    isfluid : boolean
        True if the cylinder is added
        and False if the cylinder is deleted

    Examples
    --------

    the vertical canonical cylinder centered in (0, 0, 1/2)

    >>> center = [0., 0., 0.5]
    >>> v1, v2 = [1., 0., 0.], [0., 1., 0.]
    >>> w = [0., 0., 1.]
    >>> CylinderTriangle(center, v1, v2, w)
    +------------------+
    | CylinderTriangle |
    +------------------+
        - dimension: 3
        - center: [0.  0.  0.5]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0. 0. 1.]
        - label: [0, 0, 0, 0, 0]
        - type: solid

    """
    def __init__(self, center, v1, v2, w, label=0, isfluid=False):
        self.number_of_bounds = 5  # number of edges
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        self.w = np.asarray(w)
        self.change_of_variables()
        self.base = BaseTriangle(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)


class Parallelepiped(Cylinder):
    """
    Class Parallelepiped

    Parameters
    ----------
    point : list
        the three coordinates of the first point
    v0 : list
        the three coordinates of the first vector that defines the edge
    v1 : list
        the three coordinates of the second vector that defines the edge
    v2 : list
        the three coordinates of the third vector that defines the edge
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Attributes
    ----------
    number_of_bounds : int
        6
    dim: int
        3
    point : ndarray
        the coordinates of the first point of the parallelepiped
    v0 : list
        the three coordinates of the first vector
    v1 : list
        the three coordinates of the second vector
    v2 : list
        the three coordinates of the third vector
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the parallelepiped is added
        and False if the parallelepiped is deleted

    Examples
    --------

    the vertical canonical cube centered in (0, 0, 0)

    >>> center = [0., 0., 0.5]
    >>> v0, v1, v2 = [1., 0., 0.], [0., 1., 0.], [0., 0., 1.]
    >>> Parallelepiped(center, v0, v1, v2)
    +----------------+
    | Parallelepiped |
    +----------------+
        - dimension: 3
        - center: [0. 0. 1.]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0.  0.  0.5]
        - label: [0, 0, 0, 0, 0, 0]
        - type: solid

    """
    def __init__(self, point, v0, v1, v2, label=0, isfluid=False):
        self.number_of_bounds = 6  # number of edges
        self.point = np.asarray(point)
        self.v1 = np.asarray(v0)
        self.v2 = np.asarray(v1)
        self.w = .5*np.asarray(v2)
        self.center = self.point + self.w
        self.change_of_variables()
        self.base = BaseParallelogram(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
A flexible Python package for lattice Boltzmann method.
"""

import logging
from colorlog import ColoredFormatter
import mpi4py.MPI as mpi
from colorama import init
init()

# pylint: disable=wrong-import-position
# pylint: disable=wildcard-import
# pylint: disable=invalid-name

from .version import version as __version__  # noqa: E402
from .domain import Domain                   # noqa: E402
from .stencil import Stencil                 # noqa: E402
from .simulation import Simulation           # noqa: E402
from . import boundary as bc                 # noqa: E402
from .scheme import Scheme                   # noqa: E402
from .elements import *                      # noqa: E402
from .geometry import Geometry               # noqa: E402
from . import viewer                         # noqa: E402
from .hdf5 import H5File                     # noqa: E402
from .options import options                 # noqa: E402
from . import monitoring                     # noqa: E402
from .analysis import EquivalentEquation, Stability  # noqa: E402
from .utils import progress_bar              # noqa: E402

numeric_level = getattr(logging, options().loglevel, None)

formatter = ColoredFormatter(
        f"%(log_color)s[{mpi.COMM_WORLD.Get_rank()}] "
        "%(levelname)-8s %(name)s in function %(funcName)s "
        "line %(lineno)s\n%(reset)s%(message)s",
        datefmt=None,
        reset=True,
        log_colors={
                'DEBUG':    'green',
                'INFO':     'cyan',
                'WARNING':  'blue',
                'ERROR':    'red',
                'CRITICAL': 'red,bg_white',
        },
        style='%'
)

logger = logging.getLogger(__name__)
logger.setLevel(level=numeric_level)

console = logging.StreamHandler()
console.setFormatter(formatter)
logger.addHandler(console)



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
__init__ of the module containing the geometrical elements
"""

from .circle import Circle
from .ellipse import Ellipse
from .parallelogram import Parallelogram
from .triangle import Triangle

from .sphere import Sphere
from .ellipsoid import Ellipsoid
from .cylinder import CylinderCircle, CylinderEllipse, CylinderTriangle
from .cylinder import Parallelepiped

from .stl_element import STLElement

__all__ = ['Circle', 'Ellipse', 'Parallelogram', 'Triangle',
           'Sphere', 'Ellipsoid', 'CylinderCircle', 'CylinderEllipse', 'CylinderTriangle', 'Parallelepiped', 'STLElement']


# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Storage module
"""

import copy
import logging
from six.moves import range

import numpy as np
import sympy as sp
import mpi4py.MPI as mpi

from .generator import For
from .monitoring import monitor

log = logging.getLogger(__name__) # pylint: disable=invalid-name


class Array:
    """
    This class defines the storage of the moments and
    distribution functions in pylbm.

    It sets the storage in memory and how to access.

    Parameters
    ----------
    nv: int
        number of velocities
    gspace_size: list
        number of points in each direction including the fictitious point
    vmax: list
        the size of the fictitious points in each direction
    sorder: list
        the order of nv, nx, ny and nz
        Default is None which mean [nv, nx, ny, nz]
    mpi_topo: MpiTopology
        the mpi topology
    dtype: type
        the type of the array. Default is numpy.double
    gpu_support : bool
        true if GPU is needed

    Attributes
    ----------
    array
    nspace
    nv
    shape
    size

    """
    #pylint: disable=too-many-locals
    def __init__(self, nv, gspace_size, vmax, sorder=None,
                 mpi_topo=None, dtype=np.double, gpu_support=False):
        self.comm = mpi.COMM_WORLD
        self.sorder = sorder

        self.gspace_size = gspace_size
        self.dim = len(gspace_size)
        self.consm = {}
        self.gpu_support = gpu_support

        if mpi_topo is not None:
            self.mpi_topo = mpi_topo

            self.region = mpi_topo.get_region(*gspace_size)
            self.region_size = []
            for i in range(self.dim):
                self.region[i][0] -= vmax[i]
                self.region[i][1] += vmax[i]
                self.region_size.append(self.region[i][1] - self.region[i][0])
        else:
            self.region_size = gspace_size

        if sorder is None:
            ind = [i for i in range(self.dim + 1)]
        else:
            if len(sorder) != self.dim + 1:
                log.error("storage order must have the same length of the spatial dimension + 1.")
            ind = copy.deepcopy(sorder)

        self.index = copy.copy(ind)
        self.vmax = vmax

        tmpshape = [nv] + self.region_size
        shape = [0]*len(tmpshape)
        for i in range(self.dim + 1):
            shape[ind[i]] = int(tmpshape[i])
        self.array_cpu = np.zeros((shape), dtype=dtype)
        self.array = self.array_cpu

        if self.gpu_support:
            try:
                import pyopencl as cl
                import pyopencl.array #pylint: disable=unused-variable
                from .context import queue
            except ImportError:
                raise ImportError("Please install loo.py")
            self.array = cl.array.to_device(queue, self.array_cpu)

        self.swaparray = np.transpose(self.array_cpu, self.index)

        if mpi_topo is not None:
            self._set_subarray()

        # if self.gpu_support:
        #     self.generate()

    def __getitem__(self, key):
        if self.gpu_support:
            self.array_cpu[...] = self.array.get()
        if isinstance(key, sp.Symbol):
            return self.swaparray[self.consm[key]]
        return self.swaparray[key]

    def __setitem__(self, key, values):
        if isinstance(key, sp.Symbol):
            self.swaparray[self.consm[key]] = values
        else:
            self.swaparray[key] = values
        if self.gpu_support:
            try:
                import pyopencl as cl
                import pyopencl.array #pylint: disable=unused-variable
                from .context import queue
            except ImportError:
                raise ImportError("Please install loo.py")
            self.array = cl.array.to_device(queue, self.array_cpu)

    def _in(self, key):
        ind = []
        for vmax in self.vmax:
            ind.append(slice(vmax, -vmax))
        ind = np.asarray(ind)

        if self.gpu_support:
            self.array_cpu[...] = self.array.get()
        if isinstance(key, (sp.Symbol, sp.IndexedBase)):
            return self.swaparray[self.consm[key]][tuple(ind)]
        return self.swaparray[key][tuple(ind)]


    def set_conserved_moments(self, consm):
        """
        add conserved moments information to have a direct access.

        Parameters
        ----------
        consm : dict
            set the name and the location of the conserved moments.
            The format is

            key : the conserved moment (sympy symbol or string)

            value : list of 2 integers

                first item : the scheme number

                second item : the index of the conserved moment in this scheme

        """
        for k, v in consm.items():
            self.consm[k] = v

    @property
    def nspace(self):
        """
        the space size.
        """
        return self.swaparray.shape[1:]

    @property
    def nv(self):
        """
        the number of velocities.
        """
        return self.swaparray.shape[0]

    @property
    def shape(self):
        """
        the shape of the array that stores the data.
        """
        return self.array.shape

    @property
    def size(self):
        """
        the size of the array that stores the data.
        """
        return self.array.size

    #pylint: disable=too-many-locals
    def _set_subarray(self):
        """
        Create the neigbors and the subarrays to update interfaces
        between each processes.

        """
        nspace = list(self.nspace)
        nv = self.nv
        dim = self.dim
        vmax = self.vmax

        def swap(array_in):
            array_out = [0]*(dim+1)
            for i in range(dim+1):
                array_out[self.index[i]] = array_in[i]
            return array_out

        sizes = swap([nv] + nspace)

        rank = self.mpi_topo.cartcomm.Get_rank()
        coords = self.mpi_topo.cartcomm.Get_coords(rank)

        self.neighbors = []
        direction = []
        for i in range(dim):
            direction = copy.copy(coords)
            direction[i] -= 1
            self.neighbors.append(self.mpi_topo.cartcomm.Get_cart_rank(direction))
            direction = copy.copy(coords)
            direction[i] += 1
            self.neighbors.append(self.mpi_topo.cartcomm.Get_cart_rank(direction))

        self.send_tag = [0, 1, 2, 3, 4, 5]
        self.recv_tag = [1, 0, 3, 2, 5, 4]

        self.send_type = []
        self.recv_type = []

        for d in range(dim): #pylint: disable=invalid-name
            subsizes = [nv] + nspace
            subsizes[d+1] = vmax[d]
            subsizes = swap(subsizes)

            sstart = [0]*(dim+1)
            sstart[d+1] = vmax[d]
            sstart = swap(sstart)
            rstart = [0]*(dim+1)

            self.send_type.append(mpi.DOUBLE.Create_subarray(sizes, subsizes, sstart))
            self.recv_type.append(mpi.DOUBLE.Create_subarray(sizes, subsizes, rstart))

            log.info("[%d] send to %d with tag %d subarray:%s", rank, self.neighbors[2*d], self.send_tag[2*d], (sizes, subsizes, sstart))
            log.info("[%d] recv from %d with tag %d subarray:%s", rank, self.neighbors[2*d], self.recv_tag[2*d], (sizes, subsizes, rstart))

            sstart = [0]*(dim+1)
            sstart[d+1] = nspace[d] - 2*vmax[d]
            sstart = swap(sstart)
            rstart = [0]*(dim+1)
            rstart[d+1] = nspace[d] - vmax[d]
            rstart = swap(rstart)

            self.send_type.append(mpi.DOUBLE.Create_subarray(sizes, subsizes, sstart))
            self.recv_type.append(mpi.DOUBLE.Create_subarray(sizes, subsizes, rstart))

            log.info("[%d] send to %d with tag %d subarray:%s", rank, self.neighbors[2*d+1], self.send_tag[2*d+1], (sizes, subsizes, sstart))
            log.info("[%d] recv from %d with tag %d subarray:%s", rank, self.neighbors[2*d+1], self.recv_tag[2*d+1], (sizes, subsizes, rstart))

        for send, recv in zip(self.send_type, self.recv_type):
            send.Commit()
            recv.Commit()

    #pylint: disable=possibly-unused-variable
    @monitor
    def update(self):
        """
        update ghost points on the interface with the datas of the neighbors.
        """
        if self.gpu_support:
            # FIXME: move the generated code outside for loopy
            from .symbolic import call_genfunction

            dim = len(self.nspace)
            nx = self.nspace[0]
            if dim > 1:
                ny = self.nspace[1]
            if dim > 2:
                nz = self.nspace[2]
            nv = self.nv

            f = self.array

            args = locals()
            call_genfunction(self.generator.module.update_x, args)

            if dim > 1:
                call_genfunction(self.generator.module.update_y, args)
            if dim > 2:
                call_genfunction(self.generator.module.update_z, args)

        else:
            for d in range(self.dim): #pylint: disable=invalid-name
                req = []

                req.append(self.comm.Irecv([self.array, self.recv_type[2*d]], source=self.neighbors[2*d], tag=self.recv_tag[2*d]))
                req.append(self.comm.Irecv([self.array, self.recv_type[2*d + 1]], source=self.neighbors[2*d + 1], tag=self.recv_tag[2*d + 1]))

                req.append(self.comm.Isend([self.array, self.send_type[2*d]], dest=self.neighbors[2*d], tag=self.send_tag[2*d]))
                req.append(self.comm.Isend([self.array, self.send_type[2*d + 1]], dest=self.neighbors[2*d + 1], tag=self.send_tag[2*d + 1]))

                mpi.Request.Waitall(req)

    #pylint: disable=too-many-locals
    def generate(self, generator):
        """
        generate periodic conditions functions for loo.py backend.
        """
        self.generator = generator
        def set_order(array, remove_index=None):
            out = [-1]*len(self.sorder)
            for i, s in enumerate(self.sorder):
                out[s] = array[i]
            if remove_index:
                out.pop(self.sorder[remove_index])
            return out

        nx, ny, nz, nv = sp.symbols('nx, ny, nz, nv', integer=True)
        shape = set_order([nv, nx, ny, nz])

        i = sp.Idx('i', (0, nx))
        j = sp.Idx('j', (0, ny))
        k = sp.Idx('k', (0, nz))
        s = sp.Idx('s', (0, nv))

        fi = sp.IndexedBase('f', shape) #pylint: disable=invalid-name
        f_store = sp.Matrix([fi[set_order([s, 0, j, k])], fi[set_order([s, nx-1, j, k])]])
        f_load = sp.Matrix([fi[set_order([s, nx-2, j, k])], fi[set_order([s, 1, j, k])]])
        iloop = set_order([s, i, j, k], remove_index=1)
        generator.add_routine(('update_x', For(iloop, sp.Eq(f_store, f_load))))

        if len(self.sorder) > 2:
            f_store = sp.Matrix([fi[set_order([s, i, 0, k])], fi[set_order([s, i, ny-1, k])]])
            f_load = sp.Matrix([fi[set_order([s, i, ny-2, k])], fi[set_order([s, i, 1, k])]])
            iloop = set_order([s, i, j, k], remove_index=2)
            generator.add_routine(('update_y', For(iloop, sp.Eq(f_store, f_load))))

        if len(self.sorder) > 3:
            f_store = sp.Matrix([fi[set_order([s, i, j, 0])], fi[set_order([s, i, j, nz-1])]])
            f_load = sp.Matrix([fi[set_order([s, i, j, nz-2])], fi[set_order([s, i, j, 1])]])
            iloop = set_order([s, i, j, k], remove_index=3)
            generator.add_routine(('update_z', For(iloop, sp.Eq(f_store, f_load))))

class SOA(Array):
    """
    This class defines a structure of arrays to store the
    unknowns of the lattice Boltzmann schemes.

    Parameters
    ----------
    nv: int
        number of velocities
    gspace_size: list
        number of points in each direction including the fictitious point
    vmax: list
        the size of the fictitious points in each direction
    mpi_topo: MpiTopology
        the mpi topology
    dtype: type
        the type of the array. Default is numpy.double
    gpu_support: bool
        True if GPU is needed

    Attributes
    ----------
    array
    nspace
    nv
    shape
    size

    """
    def __init__(self, nv, gspace_size, vmax, mpi_topo, dtype=np.double, gpu_support=False):
        sorder = [i for i in range(len(gspace_size) + 1)]
        Array.__init__(self, nv, gspace_size, vmax, sorder, mpi_topo, dtype, gpu_support=gpu_support)

    def reshape(self):
        """
        reshape.
        """
        return self.array

class AOS(Array):
    """
    This class defines an array of structures to store the
    unknowns of the lattice Boltzmann schemes.

    Parameters
    ----------
    nv: int
        number of velocities
    gspace_size: list
        number of points in each direction including the fictitious point
    vmax: list
        the size of the fictitious points in each direction
    mpi_topo: MpiTopology
        the mpi topology
    dtype: type
        the type of the array. Default is numpy.double
    gpu_support: bool
        True if GPU is needed

    Attributes
    ----------
    array
    nspace
    nv
    shape
    size

    """
    def __init__(self, nv, gspace_size, vmax, mpi_topo, dtype=np.double, gpu_support=False):
        sorder = [len(gspace_size)] + [i for i in range(len(gspace_size))]
        Array.__init__(self, nv, gspace_size, vmax, sorder, mpi_topo, dtype, gpu_support=gpu_support)

    def reshape(self):
        """
        reshape
        """
        return self.array.reshape((np.prod(self.nspace), self.nv))



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

from tokenize import (generate_tokens, untokenize, TokenError,
    NUMBER, STRING, NAME, OP, ENDMARKER, ERRORTOKEN)
import sympy as sp
from sympy.parsing.sympy_parser import parse_expr as sympy_parse_expr
from sympy.parsing.sympy_parser import standard_transformations
from ..symbolic import set_order

def reorder(tokens, local_dict, global_dict):
    result = []
    tokens_ordered = set_order(tokens, local_dict['sorder'])
    for token in tokens_ordered:
        for t in token:
            result.append(t)
        result.append((OP, ','))
    return result

def transform_expr(tokens, local_dict, global_dict):
    result = []
    l = 0
    while l < len(tokens) - 1:
        token = tokens[l]
        nextTok = tokens[l+1]
        if token[0] == NAME and token[1] in local_dict['consm'].keys():
            if nextTok[0] == OP and nextTok[1] == '[':
                result.append((NAME, 'm'))
                result.append(nextTok)
                stacks = [[(NUMBER, '%s'%local_dict['consm'][token[1]])]]
                stacks.append([])
                l += 2
                token = tokens[l]
                while token[1] != ']':
                    if token[1] == ',':
                        stacks.append([])
                    else:
                        stacks[-1].append(token)
                    l += 1
                    token = tokens[l]
                result.extend(reorder(stacks, local_dict, global_dict))
                result.append(token)
            else:
                result.append((NAME, 'm'))
                result.append((OP, '['))
                default_index = [local_dict['consm'][token[1]]] + local_dict['default_index']
                if local_dict['sorder']:
                    index_ordered = set_order(default_index, local_dict['sorder'])
                else:
                    index_ordered = default_index
                for index in index_ordered:
                    if isinstance(index, sp.Idx):
                        result.append((NAME, '%s'%index))
                    else:
                        result.append((NUMBER, '%s'%index))
                    result.append((OP, ','))
                result.append((OP, ']'))
        else:
            result.append(token)
        l += 1
    while l < len(tokens):
        result.append(tokens[l])
        l += 1
    return result

def parse_expr(expr, user_local_dict):
    local_dict = user_local_dict
    for s in expr.atoms(sp.Symbol):
        local_dict[s.name] = s

    transformations = (standard_transformations + (transform_expr,))
    return sympy_parse_expr(expr.__str__(), local_dict=local_dict, transformations=transformations)